/*
 * 	viscontrol.h - Visualisation control header; "glue" between user interface and scene
 *	Copyright (C) 2011, D Haley 

 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.

 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.

 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef VISCONTROL_H
#define VISCONTROL_H

#include "wxcomponents.h"
#include <wx/treectrl.h>
#include <wx/grid.h>
#include <wx/listbox.h>

#include <string>
#include <vector>
#include <utility>
#include <list>
#include <stack>
#include <deque>




class VisController;
#include "scene.h"

#include "filtertree.h"
#include "plot.h"

const unsigned int MAX_UNDO_SIZE=10;


//TODO: Find a better home sourcefile for this function?
//!Update a wxTree control to layout according to the specified filter tree
void upWxTreeCtrl(const FilterTree &filterTree, wxTreeCtrl *t,
		std::map<size_t,Filter *> &filterMap,vector<const Filter *> &persistentFilters,
		const Filter *visibleFilt);


//!Visualisation controller
/*!
 * Keeps track of what visualisation controls the user has available
 * such as cameras, filters and data groups. 
 * This is essentially responsible for interfacing between program
 * data structures and the user interface.
 *
 * Only one of these should be instantiated at any time (such as due to abort mechanisms).
 */
class VisController
{
	private:
		//!Target scene
		Scene *targetScene;
		//!Target Plot wrapper system
		PlotWrapper *targetPlots;
		//!Target raw grid
		wxGrid *targetRawGrid;

		//!UI element for console output
		wxTextCtrl *textConsole;

		//!UI element for selecting plots from a list (for enable/disable)
		wxListBox *plotSelList;


		//--- Data storage ----
		//!Primary data storage for filter tree
		FilterTree filterTree;
	
		//!Undo filter tree stack 
		std::deque<FilterTree> undoFilterStack,redoFilterStack;

		//!Named, stored trees that can be put aside for secondary use
		std::vector<std::pair<std::string,FilterTree> > stashedFilters;

		//!Unique IDs for stash
		UniqueIDHandler stashUniqueIDs;
		//--------------------

		//!True if viscontrol should abort current operation
		bool doProgressAbort;
	
		//!True if viscontrol is in the middle of a refresh operation
		bool amRefreshing;

		//!True if there are pending updates from the user
		bool pendingUpdates;

		//!Have we implicitly used relative references when saving data files?
		unsigned int useRelativePathsForSave;

		//!Current progress
		ProgressData curProg;


		void clear();
	

		//!Retreive the updates to the filter tree from the scene
		void getFilterUpdates();



		//!Push the current filter tree onto the undo stack
		void pushUndoStack(); 


		//!Erase the redo stack
		void clearRedoStack();
	
		

		//!Update the console strings
		void updateConsole(const std::vector<std::string> &v, const Filter *f) const;

		//!Force an update to the scene. 
		unsigned int updateScene(std::list<vector<const FilterStreamData *> > &outputData);
		
		//!ID handler that assigns each filter its own ID that
		// is guaranteed to be unique for the life of the filter in the filterTree	
		std::map<size_t, Filter * > filterMap;		
		
		//Filters that should be able to be seen next time we show
		// the wxTree control
		std::vector<const Filter *> persistentFilters;

	public:
		VisController();
		~VisController();

		//Filter tree access functions
		//-----------------
		//!Run a refresh of the underlying tree
		//Overall progress give the progress for all filters, as a numeral (1 of n)
		//filterprogress gives the progress within a filter, as a percentage
		//curFilter gives the current filter, or the latest filter if exiting
		//on error. if == 0, then scene update has been initiated
		unsigned int refreshFilterTree(bool doUpdateScene=true);


		//obtain the outputs from the filter tree's refresh. 
		// The outputs *must* be deleted with safeDeleteFilterList
		unsigned int refreshFilterTree(
			std::list<std::pair<Filter *, 
				std::vector<const FilterStreamData * > > > &outputData); 

		//Obtain a clone of the active filter tree
		void cloneFilterTree(FilterTree &f) const{f=filterTree;};

		//!Safely delete data generated by refreshFilterTree(...). 
		//a mask can be used to *prevent* STREAM_TYPE_blah from being deleted. Deleted items are removed from the list.
		void safeDeleteFilterList(std::list<std::pair<Filter *, std::vector<const FilterStreamData * > > > &outData, 
								size_t typeMask=0, bool maskPrevents=true) const;


		//!Add a new filter to the tree. set isbase=false and parentID for not
		//setting a parent (ie makeing filter base)
		void addFilter(Filter *f, bool isBase, size_t parentId);
		
		
		//!Add a new subtree to the tree. Note that the tree will be cleared
		// as a result of this operation. Control of all pointers will be handled internally.
		// If you wish to use ::getFilterById you *must* rebuild the tree control with
		// ::updateWxTreeCtrl
		void addFilterTree(FilterTree &f,bool isBase=true, 
						size_t parentId=(unsigned int)-1); 

		//!Grab the filter tree from the internal one, and swap the 
		// internal with a cloned copy. Can be used eg, to steal the cache
		// Note that the passed filter tree will be destroyed.
		void switchoutFilterTree(FilterTree &f);

		//Perform a swap operation on the filter tree. 
		// - *must* have same topology, or you must call updateWxTreeCtrl
		void swapFilterTree(FilterTree &f) { f.swap(filterTree);}

		//!Duplicate a branch of the tree to a new position. Do not copy cache,
		bool copyFilter(size_t toCopy, size_t newParent,bool copyToRoot=false) ;
		
		const Filter* getFilterById(size_t filterId) const; 

		//!Return all of a given type of filter from the filter tree
		void getFiltersByType(std::vector<const Filter *> &filters, unsigned int type)  const;

		//!Returns true if the tree contains any state overrides (external entity referrals)
		bool hasStateOverrides() const 
				{return filterTree.hasStateOverrides();};

		//!Make the filters safe for the end user, assuming the filter tree could have had
		// its data initialised from anywhere
		void stripHazardousContents() { filterTree.stripHazardousContents();};

		//!Return the number of filters currently in the main tree
		size_t numFilters() const { return filterTree.size();};

		//!Clear the cache for the filters
		void purgeFilterCache() { filterTree.purgeCache();};

		//!Delete a filter and all its children
		void removeFilterSubtree(size_t filterId);

		//Move a filter from one part of the tree to another
		bool reparentFilter(size_t filterID, size_t newParentID);

		//!Set the properties using a key-value result (as obtained from updatewxPropertyGrid)
		/*
		 * The return code tells whether to reject or accept the change. 
		 * need update tells us if the change to the filter resulted in a change to the scene
		 */
		bool setFilterProperty(size_t filterId, unsigned int set,
				unsigned int key, const std::string &value, bool &needUpdate);
	
		//!Set the filter's string	
		bool setFilterString(size_t id, const std::string &s);
		
		//!Clear all caches
		void clearCache();
		
		//!Clear all caches
		void clearCacheByType(unsigned int type) { filterTree.clearCacheByType(type);};

		//-----------------





		//!Call to get viscontrol to abort current operation. Call once per abort.
		void abort() {ASSERT(!doProgressAbort);doProgressAbort=true;} 

		//!Call to set window to be partially excluded (wx dependant) from blocking during scene updates
		void setYieldWindow(wxWindow *win);

		//!Set the text console
		void setConsole(wxTextCtrl *t) { textConsole = t;}
		//!Set the backend scene
		void setScene(Scene *theScene);
		//!Set the backend plot
		void setPlotWrapper(PlotWrapper *thePlots){targetPlots=thePlots;};
		//!Set the listbox for plot selection
		void setPlotList(wxListBox *box){plotSelList=box;};
		
		
		//!Set the backend grid control for raw data
		void setRawGrid(wxGrid *theRawGrid){targetRawGrid=theRawGrid;};
	
	
		//!Write out the filters into a wxtreecontrol.
		void updateWxTreeCtrl(wxTreeCtrl *t,const Filter *f=0);
		//!Update a wxtGrid with the properties for a given filter
		void updateFilterPropGrid(wxPropertyGrid *g,size_t filterId);
			



		//!Set the camera to use in the scene
		bool setCam(unsigned int uniqueID) ;

		//!Remove a camera from the scene
		bool removeCam(unsigned int uniqueID);

		//!Add a new camera to the scene
		unsigned int addCam(const std::string &camName);

		//!Update a wxtGrid with the properties for a given filter
		void updateCamPropertyGrid(wxPropertyGrid *g,unsigned int camUniqueId);
		
		//!Return the number of cameras
		unsigned int numCams() const ;

		//!Set the properties using a key-value result (as obtaed from updatewxPropertyGrid)
		/*! The return code tells whether to reject or accept the change. 
		 */
		bool setCamProperties(size_t camUniqueID, 
				unsigned int key, const std::string &value);


		//!Ensure visible
		void ensureSceneVisible(unsigned int direction);

		
		//!Save the viscontrol state: writes an XML file containing the viscontrol state
		bool saveState(const char *filename, std::map<string,string> &fileMapping,
					bool writePackage=false) const;
		
		//!Save the viscontrol "package":this  writes an XML file containing the viscontrol state, altering the output of the filters to obtain the files it needs 
		bool savePackage(const char *filename) const;

		//!Load the viscontrol state.	
		bool loadState(const char *filename, std::ostream &f,bool merge=false);	

		//!Are we currently using relative paths due to a previous load?
		bool usingRelPaths() const { return useRelativePathsForSave;};

		//!Set whether to use relative paths when saving
		void setUseRelPaths(bool useRelPaths){ useRelativePathsForSave=useRelPaths;};

		//!Get the camera ID-pair data TODO: this is kinda a halfway house between
		//updating the camera data internally and passing in the dialog box and not
		//should homogenise this...
		void getCamData(std::vector<std::pair<unsigned int, std::string> > &camData) const;

		//!Get the active camera ID
		unsigned int getActiveCamId() const;

		//!export given filterstream data pointers
		unsigned int exportIonStreams(const std::vector<const FilterStreamData *> &selected, 
								const std::string &outFile, unsigned int format=IONFORMAT_POS) const;
		
		//!Returns true if the filter is in the midst of a refresh
		bool isRefreshing() const {return amRefreshing; }

		//!Inform viscontrol that it has new updates to filters from external sources (eg bindings)
		void setUpdates() { pendingUpdates=true;};

		//!Returns true if the scene has updates that need to be processed
		bool hasUpdates() const; 


		//"Stash" operations
		//	- The stashes are secondary trees that are not part of the update
		//	  but can be stored for reference (eg to recall common tree sequences)
		//----
		//!Create a new stash. Returns ID for stash
		unsigned int stashFilters(unsigned int filterID, const char *stashName);

		//!Add a stash as a subtree for the specified parent filter
		void addStashedToFilters(const Filter *parentFilter, unsigned int stashID);

		//!Delete a stash using its uniqueID
		void deleteStash(unsigned int stashID);

		//!Retreive the stash filters
		void getStashes(std::vector<std::pair<std::string,unsigned int > > &stashes) const;

		//!Retreive a given stash tree by ID
		void getStashTree(unsigned int stashId, FilterTree &) const;

		//!Get the number of stashes
		unsigned int getNumStashes() const { return stashedFilters.size();}

		//----

		void updateRawGrid() const;

		//!Set the percentage of ram to use for cache. 0 to disable
		void setCachePercent(unsigned int percent); 


		//!restore top filter tree from undo stack
		void popUndoStack(bool restorePopped=true);

		//!restore top filter tree from redo stack
		void popRedoStack();

		//!Get the size of the undo stack
		unsigned int getUndoSize() const { return undoFilterStack.size();};
		//!Get the size of the redo stack
		unsigned int getRedoSize() const { return redoFilterStack.size();};

		//!Get the current progress
		ProgressData getProgress() const { return curProg;}

		//!reset the progress data
		void resetProgress() { curProg.reset();};

		//!Return the scene's world axis visibility
		bool getAxisVisible();

		//!Set whether filter should use strong or weak randomisation
		void setStrongRandom(bool strongRand);

		//!Set whether to use effects or not
		void setEffects(bool enable); 

		
		//!return true if the primary tree contains hazardous filters
		bool hasHazardousContents() const ;

		//!Ask that next time we build the tree, this filter is kept visible/selected.
		//	may be used repeatedly to make more items visible, 
		//	prior to calling updateWxTreeCtrl. 
		//	filterId must exist during call.
		void setWxTreeFilterViewPersistence(size_t filterId);


		//!Ask if a given filter is a pure data source
		bool filterIsPureDataSource(size_t filterId) const ;  

#ifdef DEBUG
		//Check that the tree conrol is synced up to the filter map correctly
		void checkTree(wxTreeCtrl *t);
#endif
};


#endif
